﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reactive.Linq;
using System.Text;
using System.Threading.Tasks;
using Bosch.Core.Enumerations;
using Bosch.Cryptography;
using Bosch.Security.Sdk;
using Bosch.Security.Sdk.Dtos;
using Bosch.Security.Sdk.Enumerations;
using Bosch.Security.Sdk.Messages;
using Bosch.Security.Sdk.Interfaces;
using Bosch.TransportProvider;
using GalaSoft.MvvmLight;
using GalaSoft.MvvmLight.CommandWpf;

namespace SimpleWpfClient
{
    /// <summary>
    /// This is a simple demo that uses the SDK.
    /// </summary>
    /// <seealso cref="GalaSoft.MvvmLight.ViewModelBase" />
    public class MainWindowViewModel : ViewModelBase
    {
        private ConnectionStates _connectionState = ConnectionStates.Disconnected;
        public List<ProductIds> PanelTypes { get; set; }

        public MainWindowViewModel()
        {
            //Instantiate the SessionManager
            _sessionManager = new SessionManager(new TransportProviderFactory(), new AesCrypto());
            Bus = _sessionManager.MessageBus;

            //Register for Message Bus Events
            RegisterForMessageBusEvents();
            _requestedEvents = new List<RequestedEvent>
            {
                new RequestedEvent { PanelEvent = PanelEvents.OutputStateChanges, IntervalInSeconds = 5, MetadataRetrievalSeconds = 60 }
            };

            PanelTypes = new List<ProductIds> { ProductIds.B3512, ProductIds.B4512, ProductIds.B5512, ProductIds.B6512, ProductIds.B8512, ProductIds.B9512 };
            SelectedPanelType = PanelTypes.Last();
            _sessionId = -1;
        }

        private void RegisterForMessageBusEvents()
        {
            #region Connection events

            Bus.GetEvent<SessionConnectedEvent>()
                .Where(x => x.SessionId == _sessionId)
                .Subscribe(y =>
                {
                    LogMessage("Connected");
                    App.Current.Dispatcher.Invoke(() => ConnectionState = ConnectionStates.Connected);
                });

            Bus.GetEvent<SessionConnectFailedEvent>().Where(x => x.SessionId == _sessionId)
                .Subscribe(y =>
                {
                    LogMessage($"Connection attempt failed. {y.ActionResult}: {y.Message}");
                    ConnectionState = ConnectionStates.Disconnected;
                });

            Bus.GetEvent<SessionClosedEvent>().Where(x => x.SessionId == _sessionId)
                .Subscribe(async y =>
                {
                    ConnectionState = ConnectionStates.Disconnected;
                    if (y.SessionClosedReason == SessionClosedReasons.UserDisconnected)
                    {
                        LogMessage("User disconnected.");
                        return;
                    }
                    LogMessage("Connection lost. Reconnecting.");
                    await Task.Delay(500);
                    await _sessionManager.ConnectAutomationSession(_sessionId, _requestedEvents, false);
                });

            #endregion

            #region Output events

            Bus.GetEvent<OutputAddedEvent>()
                .Where(x => x.SessionId == _sessionId)
                .Subscribe(y =>
                {
                    string state = y.IsOn ? "On" : "Off";
                    LogMessage($"Output {y.OutputNumber} added; State: {state}");
                });

            Bus.GetEvent<OutputStateChangedEvent>()
                .Where(x => x.SessionId == _sessionId)
                .Subscribe(y =>
                {
                    string state = y.IsOn ? "On" : "Off";
                    LogMessage($"Output {y.OutputNumber} changed; State: {state}");
                });

            #endregion
        }

        private async Task ConnectToPanel()
        {
            ConnectionState = ConnectionStates.Connecting;
            SdkProfile sdkProfile = new SdkProfile
            {
                ProductId = SelectedPanelType,
                Mode2ConnectionUser = Mode2ConnectionUsers.Automation,
                PanelIpAddress = PanelAddress,
                PanelPort = Port.Value,
                ConnectionPasscode = AutomationPasscode
            };

            //Initialize the Session
            LogMessage("Initializing Session");
            _sessionId = _sessionManager.InitializeSession(sdkProfile, null);

            //Connect to the Panel
            LogMessage($"Connecting Session {_sessionId} @ {PanelAddress}:{Port}");
            await _sessionManager.ConnectAutomationSession(_sessionId, _requestedEvents, false);
            //We will get either a SessionConnectedEvent or SessionConnectFailedEvent. See RegisterForMessageBusEvents.
        }

        private void DisconnectFromPanel()
        {
            ConnectionState = ConnectionStates.Disconnecting;

            //Disconnecting from the Panel
            LogMessage($"Disconnecting Session {_sessionId} @ {PanelAddress}:{Port}");
            _sessionManager.DisconnectSession(_sessionId);

            //Destroying the Session
            LogMessage($"Destroying Session {_sessionId} @ {PanelAddress}:{Port}");
            _sessionManager.DestroySession(_sessionId);
            ConnectionState = ConnectionStates.Disconnected;
        }

        private void ControlOutput(int outputNumber, OutputCommands outputCommand)
        {
            //Controlling the Panel
            Bus.Publish(new ControlOutputCommand(_sessionId, outputNumber, outputCommand));
        }

        public IMessageBus Bus { get; set; }

        private void LogMessage(string message)
        {
            _stringBuilder.AppendLine($"{DateTime.Now:HH:mm:ss.fff} - {message}");
            AppLog = _stringBuilder.ToString();
        }

        #region Relay commands

        public RelayCommand ConnectCommand
        {
            get
            {
                return _connectCommand ?? (_connectCommand = new RelayCommand(
                    async () => { await Task.Run(ConnectToPanel); },
                    () => CanConnectToPanel));
            }
        }

        public RelayCommand DisconnectCommand
        {
            get
            {
                return _disconnectCommand ?? (_disconnectCommand = new RelayCommand(
                    DisconnectFromPanel,
                    () => ConnectionState == ConnectionStates.Connected || ConnectionState == ConnectionStates.Connecting));
            }
        }

        public RelayCommand TurnOnOutputOneCommand
        {
            get
            {
                return _turnOnOutputOneCommand ?? (_turnOnOutputOneCommand = new RelayCommand(
                    () => ControlOutput(1, OutputCommands.On),
                    () => ConnectionState == ConnectionStates.Connected || ConnectionState == ConnectionStates.Connecting));
            }
        }

        public RelayCommand TurnOffOutputOneCommand
        {
            get
            {
                return _turnOffOutputOneCommand ?? (_turnOffOutputOneCommand = new RelayCommand(
                    () => ControlOutput(1, OutputCommands.Off),
                    () => ConnectionState == ConnectionStates.Connected || ConnectionState == ConnectionStates.Connecting));
            }
        }

        #endregion

        #region Dependency properties

        public string PanelAddress
        {
            get { return _panelAddress; }
            set
            {
                if (_panelAddress == value) return;

                _panelAddress = value;
                RaisePropertyChanged();
                UpdateCanConnectToPanel();
            }
        }

        private void UpdateCanConnectToPanel()
        {
            CanConnectToPanel = _connectionState == ConnectionStates.Disconnected &&
                                !string.IsNullOrWhiteSpace(PanelAddress) && Port.HasValue &&
                                !string.IsNullOrWhiteSpace(AutomationPasscode);
        }

        public int? Port
        {
            get { return _port; }
            set
            {
                if (_port == value) return;
                _port = value;
                RaisePropertyChanged();
                UpdateCanConnectToPanel();
            }
        }

        public string AutomationPasscode
        {
            get { return _automationPasscode; }
            set
            {
                if (_automationPasscode == value) return;
                _automationPasscode = value;
                RaisePropertyChanged();
                UpdateCanConnectToPanel();
            }
        }

        public ProductIds SelectedPanelType
        {
            get { return _selectedPanelType; }
            set
            {
                if (_selectedPanelType == value) return;
                _selectedPanelType = value;
                RaisePropertyChanged();
            }
        }

        public ConnectionStates ConnectionState
        {
            get { return _connectionState; }
            set
            {
                if (_connectionState == value) return;
                _connectionState = value;
                RaisePropertyChanged();
            }
        }

        public bool CanConnectToPanel
        {
            get { return _canConnectToPanel; }
            set
            {
                if (_canConnectToPanel == value) return;
                _canConnectToPanel = value;
                RaisePropertyChanged();
            }
        }

        public string AppLog
        {
            get { return _appLog; }
            set
            {
                if (_appLog == value) return;
                _appLog = value;
                RaisePropertyChanged();
            }
        }

        #endregion

        private string _panelAddress;
        private readonly SessionManager _sessionManager;
        private int _sessionId;
        private readonly List<RequestedEvent> _requestedEvents;
        private readonly StringBuilder _stringBuilder = new StringBuilder();
        private string _appLog;
        private int? _port;
        private string _automationPasscode;
        private ProductIds _selectedPanelType;
        private RelayCommand _disconnectCommand;
        private RelayCommand _connectCommand;
        private RelayCommand _turnOnOutputOneCommand;
        private RelayCommand _turnOffOutputOneCommand;
        private bool _canConnectToPanel;
    }
}